#!/usr/bin/python

"""
wgrun.py -- run the webget queue
azz@gnu.org

This needs a Python with threads enabled; everything else is standard.

--

Modification log:
2000/01/02  altered startsome to pick random rather than sequential
            URLs to start; this should improve performance, as one
            slow site can stop any others from being started otherwise
2000/07/18  made the open() call set mode 664, so that we can use
            a group for webget
2000/07/21  fixed the above change, and removed self.base in favour
            of the constants declared in wgtools

"""

import sys
import os
import string
import thread
import urllib
import time
import whrandom
import wgtools

class webgetter:

    def __init__(self):
        # max URLs to get at once
        self.maxthreads = 20

        # time in seconds after which a remote server is assumed dead
        self.deadtime = 180

        # globals
	self.threadinit = {}
        self.threadstarted = {}
        self.threadtotal = {}
        self.threadlast = {}
        self.threadlength = {}
        self.failedurl = {}
        self.modifylock = thread.allocate_lock()
        self.queuelock = thread.allocate_lock()
        self.count = 0
        self.generator = whrandom.whrandom()

    def main(self):
        self.startsome()
        
        while self.count:
            time.sleep(3)
            self.modifylock.acquire()
            print "Status: " + str(self.count) + "/" + str(self.maxthreads)
            for x in self.threadstarted.keys():
                running = self.threadlast[x] - self.threadstarted[x]
                # junk excessively small values
                if running < 1:
                    running = 0
                n = string.ljust(x, 40)[:40]
                if self.threadstarted[x] and running <> 0:
                    rate = self.threadtotal[x] / running
                    if self.threadlength[x] > 0:
                        p = "/%-5.1fK %3d%%" % (self.threadlength[x]/1000, 100*self.threadtotal[x]/self.threadlength[x])
                    else:
                        p = ""
                    print "%s  time: %-4dS rate: %-5.1fKps got:%-5.1fK%s" % (n, running, rate/1000, self.threadtotal[x]/1000, p)
                else:
                    rate = 0
                    starting = time.time() - self.threadinit[x]
                    print "%s  time: %-4dS connecting..." % (n, starting)
            print ""
            self.modifylock.release()
            if self.count < self.maxthreads:
                self.startsome()
        print "All threads finished and nothing more to do. Goodbye."

    def startsome(self):
        self.queuelock.acquire()
        dir = os.listdir(wgtools.queuedir)
        tried = {}
        while self.count < self.maxthreads:
            i = self.generator.randint(0, len(dir) - 1)
            url = string.replace(dir[i], "@", "/")
            self.modifylock.acquire()
            y = self.threadstarted.has_key(url) or self.failedurl.has_key(url) or tried.has_key(url)
            self.modifylock.release()
            if not y:
                thread.start_new_thread(self.graburl, (url, ""))
                self.count = self.count + 1
            tried[url] = 1
            if len(tried) == len(dir):
                break
        self.queuelock.release()

    def graburl(self, url, splat):
        self.modifylock.acquire()
        self.threadinit[url] = time.time()
	self.threadstarted[url] = 0
        self.threadtotal[url] = 0
        self.threadlast[url] = self.threadstarted[url]
        self.threadlength[url] = 0
        self.modifylock.release()

        print "Started: " + url
        try:
            f = urllib.urlopen(url)
        except IOError:
            print "Failed (open): " + url
            self.modifylock.acquire()
            self.failedurl[url] = 1
            del self.threadstarted[url]
            self.count = self.count - 1
            self.modifylock.release()
            thread.exit()

        # Attempt to get the size of whatever it is we're downloading; this can go wrong
        # in a variety of interesting ways, so we wrap it in a try/except pair.
        try:
            l = string.atoi(f.info()["Content-Length"])            
        except:
            l = 0
        self.threadlength[url] = l

	self.threadstarted[url] = time.time()
	print "Connected: " + url

        data = "something"
        got = ""
        while data:
            # Occasionally read on a URL object will throw an IOError. No idea why.
            # We treat it like a failed URL (i.e. won't continue this session).
            try:
                data = f.read(512)
            except IOError:
                print "Failed (read): " + url
                self.modifylock.acquire()
                self.failedurl[url] = 1
                del self.threadstarted[url]
                self.count = self.count - 1
                self.modifylock.release()
                thread.exit()
            got = got + data
            self.modifylock.acquire()
            self.threadtotal[url] = len(got)
            self.threadlast[url] = time.time()
            self.modifylock.release()

        f.close()

	if len(got) > 0:
            # Write out the data
            name = wgtools.gotdir + wgtools.filenameof(url)
            f = open(name, "w")
            f.write(got)
            f.close()
            os.chmod(name, 0664)

            # Remove it from the queue
            self.queuelock.acquire()
            os.unlink(wgtools.queuedir + wgtools.filenameof(url))
            self.queuelock.release()

	# Note that if the length was 0 it can immediately start it
	# again. This could cause problems if the file really is of
	# length zero, but I'm not going to lose any sleep about it.

        print "Finished: " + url
        self.modifylock.acquire()
        del self.threadstarted[url]
        self.count = self.count - 1
        self.modifylock.release()
        thread.exit()

#---

# Start the fun.
o = webgetter()
o.main()

